import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.awt.geom.*;
import java.math.*;




/*The main purpose of this class is that it checks if
two triangles intersect, and then produces a point in the
intersection if they do actually intersect*/

public class Facet {
    Complex[] z=new Complex[3];

    public Facet() {}
    public Facet(Complex z1,Complex z2,Complex z3) {
	z[0]=new Complex(z1);
	z[1]=new Complex(z2);
	z[2]=new Complex(z3);
    }


    public GeneralPath toPath() {

	GeneralPath gp=new GeneralPath();
	gp.moveTo((float)(z[0].x),(float)(z[0].y));
	gp.lineTo((float)(z[1].x),(float)(z[1].y));
	gp.lineTo((float)(z[2].x),(float)(z[2].y));
	gp.closePath();
	return(gp);
    }

  /*checks if the segment z1z2 crosses the segment z3,z4*/
    public static int checkCross(Complex z1,Complex z2,Complex z3,Complex z4) {
	double a1=Complex.signedArea(z1,z2,z3);
	double a2=Complex.signedArea(z1,z2,z4);
	if(Math.abs(a1)<.0000001) return(0);
	if(Math.abs(a2)<.0000001) return(0);
	if(a1*a2>0) return(0);
        a1=Complex.signedArea(z3,z4,z1);
	a2=Complex.signedArea(z3,z4,z2);
  	if(Math.abs(a1)<.0000001) return(0);
	if(Math.abs(a2)<.0000001) return(0);
	if(a1*a2>0) return(0);
      return(1);
    }



    //1 means yes

    public static int testCross(Facet F1,Facet F2) {

	for(int i=0;i<3;++i) {
	    for(int j=0;j<3;++j) {
		int ii=(i+1)%3;
		int jj=(j+1)%3;
		Complex z1=F1.z[i];
		Complex z2=F1.z[ii];
		Complex z3=F2.z[j];
		Complex z4=F2.z[jj];
		int test=checkCross(z1,z2,z3,z4);
		if(test==1) return(1);
	    }
	}
	return(0);
    }

    public static int testContain(Complex z,Facet F) {
	double a1=Complex.signedArea(z,F.z[0],F.z[1]);
	double a2=Complex.signedArea(z,F.z[1],F.z[2]);
	double a3=Complex.signedArea(z,F.z[2],F.z[0]);
	if(a1*a2<0) return(0);
	if(a2*a3<0) return(0);
	if(a3*a1<0) return(0);
	return(1);
    }

    public static Facet correctTinyErrors(Facet F1,Facet F2) {
	Facet F3=new Facet();
	for(int i=0;i<3;++i) F3.z[i]=new Complex(F2.z[i]);
	for(int i=0;i<3;++i) {
	    for(int j=0;j<3;++j) {
		double test=Complex.dist(F1.z[i],F2.z[j]);
		if(test<.000000001) F3.z[j]=F1.z[i];
	    }
	}
	return(F3);
    }

    public static int testContain(Facet F1,Facet F2) {
	Facet F3=correctTinyErrors(F1,F2);
	for(int i=0;i<3;++i) {
	    if(testContain(F1.z[i],F3)==0) return(0);
	}
	return(1);
    }


    public static int testDisjoint(Facet F1,Facet F2) {
	if(testDegenerate(F1)==1) return(1);
	if(testDegenerate(F2)==1) return(1);
	if(edgeDisjointTest(F1,F2)==1) return(1);
	if(edgeDisjointTest(F1,F2)==2) return(0);
	if(testContain(F1,F2)==1) return(0);
	if(testContain(F2,F1)==1) return(0);
	if(testCross(F1,F2)==1) return(0);
	return(1);
    }


    /**test for degeneracy*/

    public static int testDegenerate(Facet F) {
	double a=Complex.signedArea(F.z[0],F.z[1],F.z[2]);
	if(Math.abs(a)<.0000001) return(1);
	return(0);
    }

    /**test for equality*/
    public static int testEquals(Facet F1,Facet F2) {

	int[][] index=testEdgeShare(F1,F2);
	if(index[2][0]==-1) return(0);
	return(1);
    }



    /**when the faces share an edge**/
    public static int[][] testEdgeShare(Facet F1,Facet F2) {
	int count=0;
	int[][] index=new int[5][2];
	for(int i=0;i<3;++i) {
	    for(int j=0;j<2;++j) {
              index[i][j]=-1;
	    }
	}


	for(int i=0;i<3;++i) {
	    for(int j=0;j<3;++j) {
		if(Complex.dist(F1.z[i],F2.z[j])<.00000001) {
		    index[count][0]=i;
		    index[count][1]=j;
		    ++count;
		}
	    }
	}
	return(index);
    }

    //1 means shares an edge
    public static int edgeShare(Facet F1,Facet F2) {
	int[][] index=testEdgeShare(F1,F2);
	if(index[1][0]==-1) return(0);
	return(1);
    }

    //1 means shares edge but otherwise disjoint
    //2 means shares edge but not disjoint

    public static int edgeDisjointTest(Facet F1,Facet F2) {
	int[][] index=testEdgeShare(F1,F2);
	if(index[1][0]==-1) return(0);

	int ind1=3-index[0][0]-index[1][0];
	int ind2=3-index[0][1]-index[1][1];

	Complex z1=F1.z[index[0][0]];
	Complex z2=F1.z[index[1][0]];
	Complex z3=F1.z[ind1];
	Complex w3=F2.z[ind2];

	double a1=Complex.signedArea(z1,z2,z3);
	double a2=Complex.signedArea(z1,z2,w3);
	if(a1*a2<0) return(1);
	return(0);
    }







    /**this is used when F1 is contained in F2**/

    public static Complex goodIntersection1(Facet F1,Facet F2) {
	Complex z=new Complex();
	for(int i=0;i<3;++i) {
	    z.x=z.x+F1.z[i].x/3.0;
	    z.y=z.y+F1.z[i].y/3.0;
	}
	return(z);
    }


    public static Complex goodIntersection2(Facet F1,Facet F2) {
	Complex z=new Complex();
	int count=0;
	for(int i=0;i<3;++i) {
	    for(int j=0;j<3;++j) {
		int ii=(i+1)%3;
		int jj=(j+1)%3;
		Complex z1=F1.z[i];
		Complex z2=F1.z[ii];
		Complex z3=F2.z[j];
		Complex z4=F2.z[jj];
		int test=checkCross(z1,z2,z3,z4);
		if(test==1) {
		    Complex w=findCross(z1,z2,z3,z4);
		    z=Complex.plus(z,w);
		    ++count;
		}
	    }
	}
	z.x=z.x/count;
	z.y=z.y/count;
	return(z);
    }


    /**This is used when F1 crosses F2.   We find all intersection
       points of the edges and take their convex hull. **/



    public static Complex goodIntersection3(Facet F1,Facet F2) {
	Complex z=new Complex();
	int count=0;
	for(int i=0;i<3;++i) {
	    for(int j=0;j<3;++j) {
		int ii=(i+1)%3;
		int jj=(j+1)%3;
		Complex z1=F1.z[i];
		Complex z2=F1.z[ii];
		Complex z3=F2.z[j];
		Complex z4=F2.z[jj];
		int test=checkCross(z1,z2,z3,z4);
		if(test==1) {
		    Complex w=findCross(z1,z2,z3,z4);
		    z=Complex.plus(z,w);
		    ++count;
		}
	    }
	}
	z.x=z.x/count;
	z.y=z.y/count;


	Complex w=new Complex();
	count=0;
	for(int i=0;i<3;++i) {
	    if(testContain(F1.z[i],F2)==1) {
		++count;
		w=Complex.plus(w,F1.z[i]);
	    }
	}

	for(int i=0;i<3;++i) {
	    if(testContain(F2.z[i],F1)==1) {
		++count;
		w=Complex.plus(w,F2.z[i]);
	    }
	}
	if(count==0) return(z);
	w.x=w.x/count;
	w.y=w.y/count;
	z.x=.5*(z.x+w.x);
	z.y=.5*(z.y+w.y);

	return(z);
    }




    /*This is only used when the facets are known to cross */

    public static Complex goodIntersection(Facet F1,Facet F2) {
	if(testContain(F1,F2)==1) return(goodIntersection1(F1,F2));
	if(testContain(F2,F1)==1) return(goodIntersection1(F2,F1));
	if(edgeShare(F1,F2)==1) return(goodIntersection2(F1,F2));
	return(goodIntersection3(F1,F2));
    }

    public static Complex findCross(Complex z1,Complex z2,Complex z3,Complex z4) {
	Vector v1=new Vector(z1);
	Vector v2=new Vector(z2);
	Vector v3=new Vector(z3);	
        Vector v4=new Vector(z4);
	Vector v5=Vector.findCross(v1,v2,v3,v4);
	return(v5.toComplex());
    }


    /**this is used when z is inside F*/

    public static double[] barycentricCoords(Complex z,Facet F) {
	double[] t=new double[3];
	t[2]=Complex.signedArea(z,F.z[0],F.z[1]);
	t[0]=Complex.signedArea(z,F.z[1],F.z[2]);
	t[1]=Complex.signedArea(z,F.z[2],F.z[0]);

	for(int i=0;i<3;++i) t[i]=Math.abs(t[i]);
	double s=t[0]+t[1]+t[2];
	for(int i=0;i<3;++i) t[i]=t[i]/s;
	return(t);
    }

    public static Complex fromBarycenter(double[] t,Facet F) {
	Complex z=new Complex();
	for(int i=0;i<3;++i) {
	    z.x=z.x+t[i]*F.z[i].x;
	    z.y=z.y+t[i]*F.z[i].y;
	}
	return(z);
    }


    /*This is only used when the facets are known to cross */

    public static double[] goodBarycentricCoords(Facet F1,Facet F2) {
	double[] t=new double[6];
	Complex z=goodIntersection(F1,F2);
	double[] t1=barycentricCoords(z,F1);
	double[] t2=barycentricCoords(z,F2);
	t[0]=t1[0];
	t[1]=t1[1];
	t[2]=t1[2];
	t[3]=t2[0];
	t[4]=t2[1];
	t[5]=t2[2];
	return(t);
    }


}

